/*********************************************************************
* MOVEKEYB.C - set up a larger ring buffer for the keyboard buffer.  *
*                                                                    *
* To compile: cl MOVEKEYB.C MCB.C             Last revised, 10/21/91 *
*********************************************************************/
#include <stdio.h>      /* printf(), puts()                     */
#include <stdlib.h>     /* exit(), EXIT_FAILURE                 */
#include <dos.h>        /* intdosx(), REGS, SREGS               */
#include <assert.h>     /* assert()                             */
#include "mcb.h"        /* Memory control block definitions     */

#define KEYB_HEAD   (unsigned int far *)0x0040001A
#define KEYB_TAIL   (unsigned int far *)0x0040001C
#define KEYB_START  (unsigned int far *)0x00400080
#define KEYB_END    (unsigned int far *)0x00400082

#define MAXSIZE 16      /* Maximum buffer size: 128 keystrokes  */
#define MINSIZE 3       /* Minimum buffer size: 24 keystrokes   */
#define FREE    0       /* Owner tag for free block             */
#define TRUE    1       /* Boolean true value                   */
#define FALSE   0       /* Boolean false value                  */

void main(void)
    {
    unsigned int i;            /* Counter                           */
    int found;                 /* Flag: TRUE if new buffer found    */
    MCB huge *MCBptr;          /* Pointer to current MCB in chain   */
    unsigned int far *EnvPtr;  /* Pointer to our environment block  */
    unsigned int far *HeadOff = KEYB_HEAD;
    unsigned int far *TailOff = KEYB_TAIL;
    unsigned int far *StartOff = KEYB_START;
    unsigned int far *EndOff = KEYB_END;

    assert( sizeof(MCB)==mcbSIZE );  /* Verify structure packing    */
    puts( "MOVEKEYB - Extend the size of the keyboard buffer" );

    /*----Free this program's environment block---------------------*/
    EnvPtr = MK_FP( _psp, 0 );
    printf( "Environment block address: %04x:0000, length: %d\n",
             EnvPtr[22]-1, ((MCB *)MK_FP(EnvPtr[22]-1,0))->size * 16 );
    regs.x.ax = 0x4900;
    sregs.es = EnvPtr[22];
    intdosx( &regs, &regs, &sregs );

    /*----Try to find an MCB large enough for the keyboard buffer---*/
    for ( found=FALSE, MCBptr=getFirstMCB();
          !found && MCBptr->marker=='M'; MCBptr=getNextMCB(MCBptr) )
        {
        if ( MCBptr->owner != FREE )    /* Examine only free RAM*/
            continue;
        if ( MCBptr->size > MAXSIZE )   /* Don't waste HUGE blks*/
            continue;
        if ( MCBptr->size < MINSIZE )   /* Don't use tiny blocks*/
            continue;
        found = TRUE;
        break;
        }

    /*----Did we find a suitable free block?------------------------*/
    if ( !found )
        {
        puts( "No suitable free RAM for new keyboard buffer" );
        exit( EXIT_FAILURE );
        }

    /*----Is the block we found close enough to segment 0x40?-------*/
    if ( FP_SEG(MCBptr) - 0x40U > 0x1000UL - MCBptr->size*16 )
        {
        puts( "No suitable RAM close enough to BIOS data area" );
        exit( EXIT_FAILURE );
        }

    /*----Lay claim to the block of RAM-----------------------------*/
    MCBptr->owner = 0xffff;
    printf( "Block %Fp holds %d bytes\n", MCBptr, MCBptr->size * 16 );

    /*----Set up BIOS's keyboard pointers---------------------------*/
    printf( "Our keyboard buffer is from 0040:%04x to 0040:%04x\n",
            (FP_SEG(MCBptr) - 0x40) * 16 + 16,
            (FP_SEG(MCBptr) - 0x40 + MCBptr->size) * 16 + 16);

    _disable();  /* Disable interrupts while we alter the pointers */
    *HeadOff = *TailOff = *StartOff = 
        (FP_SEG(MCBptr) - 0x40) * 16 + 16;
    *EndOff = (FP_SEG(MCBptr) - 0x40 + MCBptr->size) * 16 + 16;
    _enable();   /* Re-enable interrupts                           */
    }
